// Copyright (c) 2006 by the Consiglio Nazionale Ricerche (CNR) area 
// della Ricerca di Pisa
// 
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE Consiglio Nazionale Ricerche (CNR) makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.
//
//
// mac-bod.cc
// by Raffaello Secchi (raffaello.secchi@isti.cnr.it), CNR/ISTI


#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-tdma.h"
#include "cmu-trace.h"

/* Initialize Requester and Allocator*/ 
TdmaRequester::TdmaRequester():  qMon_(0)	
{
	bind("request_",&request_);
}


TdmaAllocator::TdmaAllocator()
{
	bind("delay_",&delay_);
	bind("active_node_",&active_node_);
	bind("max_slot_num_",&max_slot_num_);
	schedule_nxt_ = NULL;
}

AllocatorProportional::AllocatorProportional():	TdmaAllocator()
{
	deficit = 0;	
	slot_index = 0;
}


/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */

static class AllocatorProportionalClass : public TclClass {
public:
	AllocatorProportionalClass() : TclClass("Allocator/Proportional") {}
	TclObject* create(int, const char*const*) {
		return (new AllocatorProportional());
	}
} class_allocator_proportional;

static class AllocatorTclClass : public TclClass {
public:
	AllocatorTclClass() : TclClass("Allocator/Tcl") {}
	TclObject* create(int, const char*const*) {
		return (new AllocatorTcl());
	}
} class_allocator_tcl;

static class RequesterConstantClass : public TclClass {
public:
	RequesterConstantClass() : TclClass("Requester/Constant") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterConstant());
	}
} class_requester_constant;

static class RequesterRBDCClass : public TclClass {
public:
	RequesterRBDCClass() : TclClass("Requester/RBDC") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterRBDC());
	}
} class_requester_rbdc;

static class RequesterVBDCClass : public TclClass {
public:
	RequesterVBDCClass() : TclClass("Requester/VBDC") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterVBDC());
	}
} class_requester_vbdc;

static class RequesterTclClass : public TclClass {
public:
	RequesterTclClass() : TclClass("Requester/Tcl") {}
	TclObject* create(int, const char*const*) {
		return (new RequesterTcl());
	}
} class_requester_tcl;


/* Reinitialize Allocator and Requester objects after new nodes is created */
void TdmaAllocator::init() {
	
	requestv_ = mac_->requestv_;
	max_slot_num_ = mac_->max_slot_num_;
	tdma_schedule_nxt_ = mac_->tdma_schedule_nxt_;
	if (schedule_nxt_)
		delete [] schedule_nxt_;
	schedule_nxt_ = new int[max_slot_num_];
	active_node_ = mac_->active_node_;

	mac_->allocator_ = this;

}

void AllocatorProportional::init(){

	if (deficit) 
		delete[] deficit;
		
	active_node_ = mac_->active_node_;

	deficit = new double[active_node_];
	for(int i=0; i<active_node_; i++)
		deficit[i] = 0;	

	TdmaAllocator::init();
}

void AllocatorTcl::init(){
	
	Tcl& tcl = Tcl::instance();

	active_node_ = mac_->active_node_;

	// initialize AllocatorTcl instvars
	tcl.evalf("%s initvar",name());

	TdmaAllocator::init();
}

void TdmaRequester::init() {
	requestv_ = mac_->requestv_;
	slot_num_ = mac_->slot_num_;
	mac_->requester_ = this;
}

void RequesterTcl::init() {
	
	Tcl& tcl = Tcl::instance();

	// initialize RequesterTcl instvars
	tcl.evalf("%s initvar",name());
	
	TdmaRequester::init();
}


/* ======================================================================
   Command Interface with the interpreter 
   ====================================================================== */
int AllocatorProportional::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdma*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();
			return TCL_OK;
		}
	}
	return TdmaAllocator::command(argc, argv);
}

int AllocatorTcl::command(int argc, const char*const* argv)
{

	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdma*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();
			return TCL_OK;
		}
	}
	return TdmaAllocator::command(argc, argv);
}


int TdmaRequester::command(int argc, const char*const* argv){
	
	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdma*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();	
			return TCL_OK;
		}
		if (strcmp(argv[1], "setifq") == 0) {
			ifq_ = (Queue *)TclObject::lookup(argv[2]);
			if (!ifq_)
				return TCL_ERROR;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "setmon") == 0) {
			qMon_ = (QueueMonitor *)TclObject::lookup(argv[2]);
			if (!qMon_)
				return TCL_ERROR;
			return (TCL_OK);
		}
	}
	return TclObject::command(argc,argv);
}

int RequesterConstant::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterRBDC::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterVBDC::command(int argc, const char*const* argv)
{
	return TdmaRequester::command(argc, argv);
}

int RequesterTcl::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			mac_ = (MacTdma*) TclObject::lookup(argv[2]);
			if(mac_ == 0)
				return TCL_ERROR;
			init();	
			return TCL_OK;
		}
	}
	return TdmaRequester::command(argc, argv);
}


/* ======================================================================
   Debugging Routines
   ====================================================================== */

void TdmaAllocator::printrequests(){

	for(int i=0; i<active_node_ ; i++ )
	   printf("[%d]",requestv_[i]);
	printf("\n");

}

void TdmaAllocator::printrequest_nxt(){
	for(int i=0; i<max_slot_num_ ; i++ )
  		if (schedule_nxt_[i] == SLOT_UNALLOC)
			printf("[N]");
		else
			printf("[%d]", schedule_nxt_[i]);
	printf("\n");
}


/* Requester Main Functions */

void RequesterConstant::dorequest(){

	/* overwrite previous request */
	requestv_[slot_num_] = request_;
}

void RequesterRBDC::dorequest(){

	rate = (qMon_->barrivals() - pin)/(mac_->slot_packet_len_*mac_->num_frame_);
	
	/* overwrite previous request */
	requestv_[slot_num_] = rate+1;

	pin = qMon_->barrivals();
	tin = NOW;
}

void RequesterVBDC::dorequest(){

	buffer = qMon_->barrivals() - qMon_->bdepartures() - qMon_->bdrops();
	/* overwrite previous request */
	printf("%lf %d %d\n",NOW, slot_num_, 
	  buffer/(mac_->slot_packet_len_*mac_->num_frame_));
	requestv_[slot_num_] = buffer/(mac_->slot_packet_len_*mac_->num_frame_)+1;
}

void RequesterTcl::dorequest(){

	char wrk[1024];
	Tcl& tcl = Tcl::instance();
	
	// call the requester tcl
	sprintf(wrk,"%s dorequest",name());
	tcl.evalc(wrk);

	/* overwrite previous request */
	requestv_[slot_num_] = atoi(tcl.result());
	
}


/* Main Allocation  Functions */
void TdmaAllocator::allocation(){

	/* create new timer */
	copy_ = new AllocationDelay(this);

	/* store schedule_nxt to local memory */
	copy_->schedule_ = new int[max_slot_num_];

	for(int i=0; i<max_slot_num_; i++)
		copy_->schedule_[i] = schedule_nxt_[i];
	
	/* schedule the allocation */
	copy_->sched(delay_);
        
	mac_->trace_event("reques",0);
	mac_->trace_event("nxt",0);

}

void TdmaAllocator::schedule_copy(AllocationDelay* c){
	
	/* copy the sent schedule to local MAC */
	for(int i=0; i<max_slot_num_; i++)
		tdma_schedule_nxt_[i] = c->schedule_[i];

	/* free memory */
	delete c;
}

/* Implements the Proportional */
void AllocatorProportional::allocation() 
{
	int totreq = 0;		// sum of all requests
	int slot = 0;	
	int ns = 0;
	int si = slot_index;
	
	// find the sum of all requests
	for(int i=0; i<active_node_; i++)
	   totreq += requestv_[i];

	// add the previous deficit to current request
	if ( totreq > 0 )
	   for(int i=0; i<active_node_; i++) 
	      if (i!=slot_index)
		 deficit[i] += double(max_slot_num_*requestv_[i])/totreq;
	  
	while( slot<max_slot_num_ && ns<active_node_ ) {
	
	    while( ns<active_node_ && deficit[slot_index]<=EPS ) {
		slot_index = (slot_index + 1) % active_node_;
		ns++;
	    }

	    while( slot<max_slot_num_ && deficit[slot_index]>EPS ) {
	        deficit[slot_index]--;
	        schedule_nxt_[slot] = slot_index;
	        slot++;
	    }
	}

	deficit[si] += double(max_slot_num_*requestv_[si])/totreq;

	// the rest is not allocated
	if (slot<max_slot_num_)
	   for(int i=slot; i<max_slot_num_; i++)
	      if (deficit[si]>EPS) {
		 deficit[si]--;
		 schedule_nxt_[i] = si; 
	      } else 
	         schedule_nxt_[i] = SLOT_UNALLOC;

	/* operation common to all Allocators */
	TdmaAllocator::allocation();
}



/* Implements the Tcl Embedded allocation function */ 
void AllocatorTcl::allocation() 
{
	Tcl & tcl = Tcl::instance();

	char wrk[1024];
	
	// initialize the array
	for(int i=0; i<active_node_; i++)
		tcl.evalf("%s set requestv_(%d) %d",name(),i,requestv_[i]);

	// call Tcl scheduling function
	tcl.evalf("%s allocation", name());

	for(int i=0; i<max_slot_num_; i++){
		sprintf(wrk,"%s set schedule_(%d)",name(),i);
		tcl.evalc(wrk);
		schedule_nxt_[i] = atoi(tcl.result());
	}
	
	// operation common to all Allocators 
	TdmaAllocator::allocation();
}



