// -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
//
// Copyright (c) 1999 by the University of Southern California
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.
//
// $Header: /nfs/jade/vint/CVSROOT/ns-2/mac/mac-tdma.cc,v 1.12 2004/02/06 16:33:52 xuanc Exp $
//
// mac-tdma.cc
// by Xuan Chen (xuanc@isi.edu), ISI/USC
//
// Preamble TDMA MAC layer for single hop.
// Centralized slot assignment computing.


#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"

// #define DEBUG

//#include <debug.h>
#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-tdma.h"
#include "cmu-trace.h"

#define SET_RX_STATE(x)			\
{					\
	rx_state_ = (x);			\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state_ = (x);				\
}

double MacTdma::slot_time_ = 0;
double MacTdma::start_time_ = 0;
int MacTdma::active_node_ = 0;
int MacTdma::max_slot_num_ = 0;
int MacTdma::num_requests_ = 0;
int *MacTdma::tdma_schedule_ = NULL;
int *MacTdma::tdma_preamble_ = NULL;
int *MacTdma::tdma_schedule_nxt_ = NULL;
int *MacTdma::requestv_ = NULL;
TdmaAllocator *MacTdma::allocator_ = NULL;
int  MacTdma::num_frame_ = 1;

int MacTdma::tdma_ps_ = 0;
int MacTdma::tdma_pr_ = 0;

/* Phy specs from 802.11 */
static PHY_MIB PMIB = {
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength
};	


/* Timers */
void MacTdmaTimer::start(Packet *p, double time)
{
	Scheduler &s = Scheduler::instance();
	assert(busy_ == 0);
  
	busy_ = 1;
	paused_ = 0;
	stime = s.clock();
	rtime = time;
	assert(rtime >= 0.0);
 	
       	
	s.schedule(this, p, rtime);

}

void MacTdmaTimer::stop(Packet *p) 
{

	Scheduler &s = Scheduler::instance();
	assert(busy_);
  
	if(paused_ == 0)
		s.cancel((Event *)p);

	// Should free the packet p.
	Packet::free(p);
  
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
}

/* Slot timer for TDMA scheduling. */
void SlotTdmaTimer::handle(Event *e)
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->slotHandler(e);
}

/* Receive Timer */
void RxPktTdmaTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->recvHandler(e);
}

/* Send Timer */
void TxPktTdmaTimer::handle(Event *e) 
{       
	busy_ = 0;
	paused_ = 0;
	stime = 0.0;
	rtime = 0.0;
  
	mac->sendHandler(e);
}

static class MacTdmaClass : public TclClass {
public:
	MacTdmaClass() : TclClass("Mac/Tdma") {}
	TclObject* create(int, const char*const*) {
		return (new MacTdma(&PMIB));
	}
} class_mac_tdma;


// Mac Tdma definitions
// Frame format:
// Pamble Slot1 Slot2 Slot3...
MacTdma::MacTdma(PHY_MIB* p) : 
	Mac(), requester_(0), et_(0), dump_alloc_(0), 
	mhSlot_(this), mhTxPkt_(this),  mhRxPkt_(this) {

	/* Global variables setting. */
	// Setup the phy specs.
	static int slot_pointer = 0;

	phymib_ = p;

	// bind parameters of MacTdma object
	bind("slot_packet_len_", &slot_packet_len_);
	bind("max_slot_num_", &max_slot_num_);
	bind("num_frame_", &num_frame_);
	bind("dump_alloc_", &dump_alloc_);
	
	
	// Calculate the slot time based on the MAX allowed data length.
	slot_time_ = DATA_Time(slot_packet_len_);
	
	
	
	// Initialize the tdma schedule data structures.
	if ( !tdma_schedule_ ) {
		tdma_schedule_ = new int[max_slot_num_];
		tdma_preamble_ = new int[max_slot_num_];
		tdma_schedule_nxt_ = new int[max_slot_num_];
		requestv_ = new int[max_slot_num_];

		for(int i=active_node_; i<max_slot_num_; i++) {
			tdma_preamble_[i] = NOTHING_TO_SEND;
			tdma_schedule_[i] = SLOT_UNALLOC;
			tdma_schedule_nxt_[i] = SLOT_UNALLOC;
			requestv_[i] = SLOT_NOREQ;
		}
	}

	/*
	if ( active_node_ < max_slot_num_) {
		tdma_preamble_[active_node_] = active_node_;
		tdma_schedule_[active_node_] = active_node_;
		tdma_schedule_nxt_[active_node_] = active_node_;
	}
	*/

	// Record the initial active node number.
	active_node_++;
	
	/* Do each node's initialization. */
	
	// Initial channel / transceiver states. 
	tx_state_ = rx_state_ = MAC_IDLE;
	tx_active_ = 0;

	// Initialy, the radio is off. NOTE: can't use radioSwitch(OFF) here.
	radio_active_ = 0;

	// Record the start time of the new schedule.
	start_time_ = NOW;

	/* Seperate slot_num_ and the node id: 
	   we may have flexibility as node number changes.
	*/
	slot_num_ = slot_pointer++;
	
	frame_ = 0;

	if ( allocator_ )
		allocator_->init();
		
	/* Deal with preamble. */
	// Can't send anything in the first frame.
	slot_count_ = FIRST_ROUND;
	
	residue_ = 0;
	
	//Start the Slot timer...
	mhSlot_.start((Packet *) (& intr_), 0);

}

/* similar to 802.11, no cached node lookup. */
int MacTdma::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
		if (strcmp(argv[1], "eventtrace") == 0) {
			et_ = (EventTrace *)TclObject::lookup(argv[2]);
			if (!et_)
				return TCL_ERROR;
			return (TCL_OK);
		}
	}
	return Mac::command(argc, argv);
}


/* ======================================================================
   Debugging & Tracing Routines
   ====================================================================== */
void MacTdma::trace_pkt(Packet *p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_tdma* dh = HDR_MAC_TDMA(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void MacTdma::dump(char *fname)
{
	fprintf(stderr, "\n%s --- (INDEX: %d, time: %2.9f)\n", fname, 
		index_, Scheduler::instance().clock());
	
	fprintf(stderr, "\ttx_state_: %x, rx_state_: %x, idle: %d\n", 
		tx_state_, rx_state_, is_idle());
	fprintf(stderr, "\tpktTx_: %x, pktRx_: %x, callback: %x\n", 
		(int) pktTx_, (int) pktRx_, (int) callback_);
}

void MacTdma::watch_alloc(){
	
	for(int i=0; i<max_slot_num_; i++)
		if (tdma_schedule_[i] == SLOT_UNALLOC)
			printf("[N]");
		else
			printf("[%d]",tdma_schedule_[i]);
	printf("\n");	

}


void MacTdma::trace_event(char *event, int val)
{
	
	if (et_ == NULL) return;
	char *wrk = et_->buffer();
	
	if (dump_alloc_)
	   if ( strcmp(event,"req") != 0 &&
		strcmp(event,"nxt") != 0 && 
		strcmp(event,"cur") != 0 )
		return;
		
	if (wrk != 0) {
	   if (strcmp(event,"cur") == 0) {
	      static int *xc = new int[active_node_];
	      for(int i=0; i<active_node_; i++)
		   xc[i] = 0;
	      sprintf(wrk, "<%d> %lf %sslt ", index_, NOW, event);
	      for(int i=0; i<max_slot_num_; i++)
		 if (tdma_schedule_[i] == SLOT_UNALLOC)
		    sprintf(wrk+strlen(wrk), "[N]");
		 else{
		    sprintf(wrk+strlen(wrk), "[%d]", tdma_schedule_[i]);
		    xc[tdma_schedule_[i]]++;
		 }
	      sprintf(wrk+strlen(wrk), "\n<%d> %lf %sall ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++)
		   sprintf(wrk+strlen(wrk), "%d ", xc[i]);
	   } else if (strcmp(event,"reques") == 0) {
	      sprintf(wrk, "<%d> %lf %s ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++ )
		 sprintf(wrk+strlen(wrk),"%d ",requestv_[i]);
	   } else if (strcmp(event,"nxt") == 0) {
	      static int *xs = new int[active_node_];
	      for(int i=0; i<active_node_; i++)
		   xs[i] = 0;
	      sprintf(wrk, "<%d> %lf %sslt ", index_, NOW, event);
	      for(int i=0; i<max_slot_num_; i++ )
	   	if (allocator_->schedule_nxt_[i] == SLOT_UNALLOC)
		   sprintf(wrk+strlen(wrk), "[N]");
		else{
		   sprintf(wrk+strlen(wrk), "[%d]", allocator_->schedule_nxt_[i]);
		   xs[allocator_->schedule_nxt_[i]]++;
		}
	      sprintf(wrk+strlen(wrk), "\n<%d> %lf %sall ", index_, NOW, event);
	      for(int i=0; i<active_node_; i++)
		   sprintf(wrk+strlen(wrk), "%d ", xs[i]);
	   } else if (strcmp(event,"req") == 0) {
		     static double *prev = new double[active_node_];
		     static int   *prevb = new int [active_node_];
		     static int flg = 0;
		     //
		     
		     if (!flg) {
     		     	for(int i=0; i<active_node_; i++){
				prev[i] = 0.0;
				prevb[i] = 0;
		     	}
		        flg = 1;
		     }

		     int queue;
    		     
		     double dt = 0;
		     double db = 0;
		     	     
		     dt = NOW - prev[index_];
		     
		     
		     db = 8*((int)requester_->qMon_->barrivals() - 
		          (int)requester_->qMon_->bdrops()-prevb[index_]);

		     prevb[index_] = (int)requester_->qMon_->barrivals() - 
		                     (int)requester_->qMon_->bdrops();
		
		     queue = (int)requester_->qMon_->barrivals() - 
			     (int)requester_->qMon_->bdepartures() -
		             (int)requester_->qMon_->bdrops();
		     

		     sprintf(wrk, "<%d> %lf reqpar %d %lf", index_, NOW,
		   	queue,db/dt);
		     
		     prev[index_] = NOW;
		     
	   } else 
	      sprintf(wrk, "<%d> %lf %s %d", index_, NOW, event, val);
	}

	et_->trace();
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */

int MacTdma::hdr_dst(char* hdr, int dst )
{

	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(dst > -2)
		STORE4BYTE(&dst, (dh->dh_da));
	return ETHER_ADDR(dh->dh_da);
}

int MacTdma::hdr_src(char* hdr, int src )
{

	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(src > -2)
		STORE4BYTE(&src, (dh->dh_sa));
  
	return ETHER_ADDR(dh->dh_sa);
}

int MacTdma::hdr_type(char* hdr, u_int16_t type) 
{
	struct hdr_mac_tdma *dh = (struct hdr_mac_tdma*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}

/* Test if the channel is idle. */
int MacTdma::is_idle() {
	if(rx_state_ != MAC_IDLE)
		return 0;
	if(tx_state_ != MAC_IDLE)
		return 0;
	return 1;
}

/* Do slot re-scheduling */
void MacTdma::re_schedule() {

	/* make the requests */
	if (frame_ == 0 && requester_) {
		trace_event("req",slot_num_);
		requester_->dorequest();
	}

	frame_ = (frame_ + 1) % num_frame_;

	/* apply next schedule */
	for(int i=0; i<max_slot_num_; i++)
		tdma_schedule_[i] = tdma_schedule_nxt_[i];
}

/* Do slot re-allocation */
void MacTdma::re_allocate() {

	num_requests_++;

	// printf("%lf %d %d\n", NOW, num_requests_, slot_num_);

	if (num_requests_ < active_node_)
		return;

	if (allocator_)
		allocator_->allocation();

	trace_event("cur",0);

	num_requests_ = 0;
}

/* To handle incoming packet. */
void MacTdma::recv(Packet* p, Handler* h) {
	struct hdr_cmn *ch = HDR_CMN(p);
	
	u_int32_t dst;
	struct hdr_mac_tdma* dh;

	/* Incoming packets from phy layer, send UP to ll layer. 
	   Now, it is in receiving mode. 
	*/
	if (ch->direction() == hdr_cmn::UP) {
		
		dh = HDR_MAC_TDMA(p);  
		dst = ETHER_ADDR(dh->dh_da);

		// Since we can't really turn the radio off at lower level, 
		// we just discard the packet when is not directed to us
		// or is broadcast
		if (dst != (u_int32_t) index_ && dst != MAC_BROADCAST) {
			Packet::free(p);
			return;
		}

		sendUp(p);
		return;
		
	}

	/* Packets coming down from ll layer (from ifq actually),
	   send them to phy layer. 
	   Now, it is in transmitting mode. */
	
	callback_ = h;	
	state(MAC_SEND);
	sendDown(p);

}

void MacTdma::sendUp(Packet* p) 
{

	/* Detect if there is any collision happened. should not happen...?*/
	if (rx_state_ == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);     // Change the state to recv.
		pktRx_ = p;                 // Save the packet for timer reference.

		/* Start the timer for receiving, will end when receiving finishes. */
		mhRxPkt_.start(p, 0);
	} else {
		/* Note: we don't take the channel status into account, 
		   as collision should not happen...
		*/
		printf("<%d>, receiving, but the channel is not idle....???\n", index_);
	}

}

/* Actually receive data packet when RxPktTimer times out. */
void MacTdma::recvDATA(Packet *p){

	/*Adjust the MAC packet size: strip off the mac header.*/
	struct hdr_cmn *ch = HDR_CMN(p);
	// ch->size() -= ETHER_HDR_LEN;
	ch->num_forwards() += 1;

	/* Pass the packet up to the link-layer.*/
	uptarget_->recv(p, (Handler*) 0);
}

/* Send packet down to the physical layer. 
   Need to calculate a certain time slot for transmission. */
void MacTdma::sendDown(Packet* p) {
	u_int32_t dst, src, size;

	int send_byte;

	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac_tdma* dh = HDR_MAC_TDMA(p);

	/* Update the MAC header, same as 802.11 */
	// ch->size() += ETHER_HDR_LEN;

	
	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;

	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	if((u_int32_t)ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration = DATA_DURATION;
	else
		dh->dh_duration = 0;

	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();


	if (pktTx_ != NULL)
		printf("<%d> warning: buffer is NOT empty\n",index_);


	/* buffer the packet to be sent. */
	pktTx_ = p;

	/* initialize the residue to whole packet length */	
	residue_ = (HDR_CMN(pktTx_))->size();
	// printf("residue = %d\n",residue_);

	if (slot_count_ >= 0 && tdma_schedule_[slot_count_] == slot_num_ && is_idle()){ 
		send_byte = slot_packet_len_-int((NOW - slot_start_time_)*bandwidth_/8.0);
		if (residue_ <= send_byte) {
			/* the packet can be sent immediately since is my turn,
			 * the MAC is idle and enough space is in slot*/	
			tdma_preamble_[slot_count_] = (char) dst;
			send();
			residue_ = 0;
			return;
		}
		residue_ -= send_byte;
	}
}

/* Actually send the packet. */
void MacTdma::send() 
{
	u_int32_t dst, src, size;
	struct hdr_cmn* ch;
	struct hdr_mac_tdma* dh;
	double stime;

	/* Check if there is any packet buffered. */
	if (!pktTx_) {
		printf("<%d>, %f, no packet buffered.\n", index_, NOW);
		return;
	}

	/* Perform carrier sence...should not be collision...? */
	if(!is_idle()) {
		/* Note: we don't take the channel status into account, ie. no collision,
		   as collision should not happen...
		*/
		printf("<%d>, %f, transmitting, but the channel is not idle...???\n", index_, NOW);
		return;
	}

	ch = HDR_CMN(pktTx_);
	dh = HDR_MAC_TDMA(pktTx_);  

	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();
	stime = TX_Time();

	/* Turn on the radio and transmit! */
	SET_TX_STATE(MAC_SEND);						     
	radioSwitch(ON);

	/* Start a timer that expires when the packet transmission is complete. */
	mhTxPkt_.start(pktTx_->copy(), stime);
	
	downtarget_->recv(pktTx_, this);        

	pktTx_ = 0;
}

// Turn on / off the radio
void MacTdma::radioSwitch(int i) 
{
	radio_active_ = i;
	EnergyModel *em = netif_->node()->energy_model();
	if (i == ON) {
		if (em && em->sleep())
			em->set_node_sleep(0);
		//    printf("<%d>, %f, turn radio ON\n", index_, NOW);
		return;
	}

	if (i == OFF) {
		if (em && !em->sleep()) {
			em->set_node_sleep(1);
			//    netif_->node()->set_node_state(INROUTE);
		};
		//    printf("<%d>, %f, turn radio OFF\n", index_, NOW);
		return;
	}
}

// make the new preamble.
void MacTdma::makePreamble() 
{
	u_int32_t dst;
	struct hdr_mac_tdma* dh;
	
	// If there is a packet buffered, file its destination to preamble.
	if (pktTx_) {
		dh = HDR_MAC_TDMA(pktTx_);  
		dst = ETHER_ADDR(dh->dh_da);
		//printf("<%d>, %f, write %d to slot %d in preamble\n", index_, NOW, dst, slot_num_);
		tdma_preamble_[slot_count_] = dst;
	} else {
		//printf("<%d>, %f, write NO_PKT to slot %d in preamble\n", index_, NOW, slot_num_);
		tdma_preamble_[slot_count_] = NOTHING_TO_SEND;
	}
}

/* Timers' handlers */
/* Slot Timer:
   For the preamble calculation, we should have it:
   occupy one slot time,
   radio turned on for the whole slot.
*/
void MacTdma::slotHandler(Event *e) 
{
	// Restart timer for next slot.
	mhSlot_.start((Packet *)e, slot_time_);

	slot_count_++;

	// Make the new preamble for next frame.
	if ((slot_count_ == max_slot_num_) || (slot_count_ == FIRST_ROUND)) {

		/* call the requester at the beginning of frame */
		re_schedule();

		/* call the allocator at the beginning of frame */
		re_allocate();

		// We should turn the radio on for the whole slot time.
		radioSwitch(ON);

		slot_count_ = -1;
		return;
	}

	// If it is the sending slot for me.
	if (tdma_schedule_[slot_count_] == slot_num_) {
	
		slot_start_time_ = NOW;

		if (residue_ <= slot_packet_len_) {
			
			/* the packet can be sent immediately since is my turn,
			 * the MAC is idle and enough space is in slot*/	
			makePreamble();

			if (tdma_preamble_[slot_count_] != NOTHING_TO_SEND)
				send();
			else 
				radioSwitch(OFF);

			residue_ = 0;			
			return;
		}
		residue_ -= slot_packet_len_;
		return;
	}

	// If I am supposed to listen in this slot
	if ((tdma_preamble_[slot_count_] == index_) || 
		((u_int32_t)tdma_preamble_[slot_count_] == MAC_BROADCAST)) {

		// Wake up the receive packets.
		radioSwitch(ON);
		return;
	}
	
	// If I dont send / recv, do nothing.
	radioSwitch(OFF);
	return;
}

void MacTdma::recvHandler(Event *e) 
{

	u_int32_t dst, src; 
	int size;
	struct hdr_cmn *ch = HDR_CMN(pktRx_);
	struct hdr_mac_tdma *dh = HDR_MAC_TDMA(pktRx_);

	/* Check if any collision happened while receiving. */
	if (rx_state_ == MAC_COLL) 
		ch->error() = 1;

	SET_RX_STATE(MAC_IDLE);
  
	/* check if this packet was unicast and not intended for me, drop it.*/   
	dst = ETHER_ADDR(dh->dh_da);
	src = ETHER_ADDR(dh->dh_sa);
	size = ch->size();

	//printf("<%d>, %f, recv a packet [from %d to %d], size = %d\n", index_, NOW, src, dst, size);

	// Turn the radio off after receiving the whole packet
	radioSwitch(OFF);

	/* Ordinary operations on the incoming packet */
	// Not a pcket destinated to me.
	if ((dst != MAC_BROADCAST) && (dst != (u_int32_t)index_)) {
		drop(pktRx_);
		return;
	}
  
	/* Now forward packet upwards. */
	recvDATA(pktRx_);
}

/* After transmission a certain packet. Turn off the radio. */
void MacTdma::sendHandler(Event *e) 
{
	
	/* Once transmission is complete, drop the packet. 
	   p  is just for schedule a event. */
	Packet::free((Packet *)e);
	
	/* unlock IFQ. */
	SET_TX_STATE(MAC_IDLE);
	
	if(callback_) {
		Handler *h = callback_;
		callback_ = 0;
		h->handle((Event*) 0);
	}
	
	/* Turn off the radio after sending the whole packet */
	radioSwitch(OFF);
}


